open Printf
open Cil
open Slime_util
open Slime_api
open Slime_cfg
open Slime_program
open Slime_callgraph
open Slime_block
open Slime_closure
open Slime_codeloc
open Slime_verifier

let debug = if false then out else no_out

(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
let preprocess_function (fn : sac_function) : unit = begin  
  let visit_instr (instr:sac_instr) = 
    match instr with                
        (* -- Allocations -- compute static information about allocation sites. *)
      | Sac_alloc (_, size_exp, initializer_varinfo, key_exps, _) ->
          ignore (allocation_of_instr size_exp initializer_varinfo key_exps)
      | Sac_modref (_, key_exps, _) ->
          let size_exp = SizeOf(api_typ "modref_t") in
          let initializer_varinfo = api_varinfo "modref_init" in
          ignore (allocation_of_instr size_exp initializer_varinfo key_exps)
            
      (* -- Nothing. *)
      | Sac_set (def, loc) -> ()
      | Sac_call_ind (defop, funexp, exps, loc) -> () 
      | Sac_call_ord (defop, funvar, exps, loc) -> ()
      | Sac_call (funvar, args, loc) -> ()
  in
  
  let rec visit_term (term:sac_term) =
    (* -- Nothing. *)
    match term with
      | Sac_if (exp, term1, term2, loc) -> 
          ( visit_term term1 ; visit_term term2 )
      | Sac_read (def, term1, loc) -> visit_term term1
      | Sac_goto (bb_ref, loc) -> ()
      | Sac_return (expop, loc) -> ()
      | Sac_tailcall (funvar, args, loc) -> ()
  in  
  (Sac_bb_set.iter 
     (fun bb -> 
        (List.iter visit_instr bb.bb_instrs) ;
        (visit_term bb.bb_term))
     fn.fn_cfg.cfg_bbs) ;
end

(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
let elaborate_traceobjs (fn:sac_function) =
  begin match fn.fn_kind with
        
    (* Self-adjusting function. *)
    | Sac_fn_self ob -> 
        (closure_elaborate fn ob)
          
    (* Initialization function. *)
    | Sac_fn_init als ->
        (Sac_al_set.iter block_elaborate als)
          
    (* Meta function. *)
    | Sac_fn_meta -> ()
  end  

(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
let elaborate_function (cg:sac_callgraph) (fn:sac_function) =  
  let _ = debug "elaborating: %s" fn.fn_varinfo.vname in
    
  let null = CastE(voidPtrType,zero) in
  
  (* Make a new local temporary for the elaborated function. *)
  let make_temp_var (typ:typ) : varinfo =
    (makeTempVar fn.fn_fundec typ)    
  in

  let exp_of_var var = 
    Lval(Var var, NoOffset) in

  let exp_of_expop expop = 
    match expop with 
        Some exp -> exp 
      | None -> raise Hell
  in    

  (* Elaborate an optional definition into an optional lval *)
  let elab_defop (defop:sac_def option) : lval option =
    match defop with 
        None -> None
      | Some def -> Some def.df_lval
  in
    
  (* Elaborate a list of arguments -- all the readargs are elaborated
     as dereferences where the dereferenced value and modifiable being
     dereferenced are both included. *)
  let rec elab_args (noreads:bool) (args:sac_arg list) : instr list * exp list =
    match args with
        [] -> ([], [])
      | arg::args -> begin
          let (instrs, exps) = elab_args noreads args in
          match arg with
              Sac_exp(exp) -> 
                if noreads 
                then (instrs, exp::exps)
                else (instrs, exp::null::exps)

            | Sac_readarg(exp) -> 
                if noreads then invalid_arg "noreads set, but found a read" else
                let tmp = make_temp_var (TPtr(api_typ "read_t", [])) in
                let read_instr = Call 
                  (Some (Var tmp, NoOffset), 
                   exp_of_var (api_varinfo "modref_tmpread"), [exp], 
                   locUnknown) in
                (read_instr::instrs, 
                 null::(exp_of_var tmp)::exps)
        end
  in
  
  
  (* -- DEPRECATED -- *)
  (* Generate instructions to push a context for the callee if
     necessary.  Also generate instructions to restore the callers
     context, if applicable. *)
  let rec elab_callee_context_instrs (gn:sac_function) : (instr list * instr list) =
    if not (edge_is_inter_context fn gn cg) 
    then ([], []) 
    else begin
      ignore (debug "edge requires a context-switch: %s -> %s"
                fn.fn_varinfo.vname
                gn.fn_varinfo.vname) ;
      let tmp = make_temp_var (TPtr(api_typ "context_t", [])) in
      let instr1 = Call(Some (Var tmp, NoOffset), exp_of_var (api_varinfo "context_push"), [], locUnknown) in
      let instr2 = Call(None, exp_of_var (api_varinfo "context_restore"), [exp_of_var tmp], locUnknown) in
      ([instr1], [instr2])
    end
  in  

  (* Elaborate a callsite -- If a closure is necessary, returns the
     temp varinfo for the closure and instructions for before and
     after the call.  The call is "inlined" and its return value is
     bound to the temp varinfo. *)
  let elab_callsite
      (funvar:varinfo) 
      (args:sac_arg list) 
      (loc:location) 
      : (instr list * varinfo * instr list)
      =
    let (gn, ob) = lookup_self_function funvar in      
    let (arg_instrs, arg_exps) = elab_args false args in
    let codeloc_instr = elab_codeloc loc fn in
    
    (*
    let (cxt_instrs1, cxt_instrs2) = elab_callee_context_instrs gn in
    *)

    if ((List.length arg_instrs) <> 0
        || !slime_elab_closure_elim = 0)
    then 
      let tmp = make_temp_var (TPtr(ob.ob_inst_typ, [])) in
      let inst_instr = Call(Some (Var tmp, NoOffset), exp_of_var 
                              ob.ob_inst_fun, arg_exps, loc) in
      (codeloc_instr :: arg_instrs @ [inst_instr], tmp, [])
    else
      let (_, arg_exps) = elab_args true args in
      let tmp = make_temp_var (TPtr(ob.ob_inst_typ, [])) in
      let instr = Call(Some (Var tmp, NoOffset), exp_of_var funvar, arg_exps, loc) in      
      ([codeloc_instr; instr], tmp, [])
  in

  (* Elaborate a list of instructions *)
  let rec elab_instrs (instrs:sac_instr list) : instr list =
    match instrs with
        [] -> []
          
      (* Set *)
      | (Sac_set(def,loc))::instrs ->
          (Set(def.df_lval, exp_of_expop def.df_exp, loc))
          ::(elab_instrs instrs)
            
      (* Self-adjusting Call *)
      | (Sac_call(funvar, args, loc))::instrs ->
          let instrs_before, tmp, instrs_after = 
            (elab_callsite funvar args loc)
          in
          (instrs_before
           @ [Call(None, exp_of_var (api_varinfo "closure_call"), 
                   [CastE(TPtr(api_typ "closure_t", []), exp_of_var tmp)], loc)]
           @ instrs_after
           @ (elab_instrs instrs))
            
      (* Write *)
      | (Sac_call_ord(defop, funvar, exps, loc))::instrs 
          when funvar == (api_varinfo "write") ->
          let instr0 = elab_codeloc loc fn in
          (instr0
           ::(Call(elab_defop defop, exp_of_var (api_varinfo "modref_write"), exps, loc))
           ::(elab_instrs instrs))
            
      (* Ordinary Call *)
      | (Sac_call_ord(defop, funvar, exps, loc))::instrs ->
          (Call(elab_defop defop, exp_of_var funvar, exps, loc))
          ::(elab_instrs instrs)
            
      (* Indirect Call *)
      | (Sac_call_ind(defop, funexp, exps, loc))::instrs -> 
          (* BUGFIX: We want to call self-adjusting functions
             indirectly sometimes.  Since these functions (when
             normalized) will return closures, we must, as the caller,
             be prepared to trampoline these returned closures.  To
             check for this case, we use the attributes of the
             function-pointer's type to determine when to do this. *)
          let funtyp = typeOf funexp in
          let (rettyp, argtyps, isvariadic, attrs) = 
            match funtyp with 
              | TFun(rettyp, argtyps, isvariadic, attrs) ->
                  (rettyp, argtyps, isvariadic, attrs)
              | _ -> raise Hell
          in
          let attrs = (typeAttrs rettyp) @ attrs in

          (* Is the indirect call to a self-adj. function? *)
          if hasAttribute "slime_function" attrs then
            let _ = match defop with None -> () | _ -> raise Hell in
            let rettyp' = TPtr(api_typ "closure_t", []) in
            let funtyp' = TFun(rettyp', argtyps, isvariadic, attrs) in
            let funexp' = (Var (make_temp_var (TPtr(funtyp', []))), NoOffset) in
            let result  = (Var (make_temp_var rettyp'), NoOffset) in
            let instrs_out = [ 
              Set(funexp', funexp, loc) ;
              Call(Some result, Lval funexp', exps, loc) ;
              Call(None, 
                   exp_of_var (api_varinfo "closure_call"), 
                   [Lval result], loc)
            ] 
            in
            instrs_out @ (elab_instrs instrs)
          else 
            (Call(elab_defop defop, funexp, exps, loc))
            ::(elab_instrs instrs)

      (* Allocation *)
      | (Sac_alloc (def, size_exp, init_funvar, key_exps, loc))::instrs ->
          let al = allocation_of_instr size_exp init_funvar key_exps in
          let inst_fun = al.al_traceobj.ob_inst_fun in
          let tmp = make_temp_var (TPtr(al.al_traceobj.ob_inst_typ, [])) in
          let instr0 = elab_codeloc loc fn in
          let inst_args = first_n_of_list (key_exps) (count_formals inst_fun) in 
          let instr1 = Call(Some (Var tmp, NoOffset), exp_of_var inst_fun, inst_args, loc) in
          let instr2 = Call(Some def.df_lval, exp_of_var (api_varinfo "block_alloc"),
                            [CastE(TPtr(api_typ "block_t", []), exp_of_var tmp)],
                            locUnknown) in
          instr0::instr1::instr2::(elab_instrs instrs)
            
      (* Modifiable *)            
      | (Sac_modref (def, key_exps, loc))::instrs ->
          let size_exp = SizeOf(api_typ "modref_t") in
          let init_funvar = api_varinfo "modref_init" in
          let al = allocation_of_instr size_exp init_funvar key_exps in
          let inst_fun = al.al_traceobj.ob_inst_fun in
          let tmp = make_temp_var (TPtr(al.al_traceobj.ob_inst_typ,[])) in
          let instr0 = elab_codeloc loc fn in
          let instr1 = Call(Some (Var tmp, NoOffset), exp_of_var inst_fun, key_exps, loc) in
          let instr2 = Call(Some def.df_lval, exp_of_var (api_varinfo "block_alloc"),
                            [CastE(TPtr(api_typ "block_t", []), exp_of_var tmp)],
                            locUnknown) in
          instr0::instr1::instr2::(elab_instrs instrs)
  in  
  
  (* Elaborate a terminal instruction *)
  let rec elab_term 
      (bb_stmt_map:stmt Sac_bb_map.t)       
      (term:sac_term) 
      : stmt list 
      =
    match term with
        (* If *)
        Sac_if (exp, term1, term2, loc) ->
          let block1 = mkBlock (elab_term bb_stmt_map term1) in
          let block2 = mkBlock (elab_term bb_stmt_map term2) in
          [mkStmt (If (exp, block1, block2, loc))]

      | Sac_read _ -> raise Hell
          
      (* Tailcall *)
      | Sac_tailcall (funvar, args, loc) ->
          let instrs_before, tmp, instrs_after = 
            (elab_callsite funvar args loc)
          in
          let ret = CastE (TPtr(api_typ "closure_t", []), exp_of_var tmp) in
          [mkStmt (Instr (instrs_before)); 
           mkStmt (Return (Some ret, locUnknown))]
            
      (* Goto *)
      | Sac_goto (bb_ref, loc) ->
          let bb_stmt = Sac_bb_map.find !bb_ref bb_stmt_map in
          [mkStmt (Goto(ref bb_stmt, loc))]
            
      (* Return *)
      | Sac_return (expop, loc) -> begin
          match (fn.fn_kind, expop) with
            | (Sac_fn_self _, None) -> [mkStmt (Return(Some null, loc))]
            | (Sac_fn_self _, _) -> raise Hell
            | (Sac_fn_init _, None) -> [mkStmt (Return(None, loc))]
            | (Sac_fn_init _, _) -> raise Hell
            | (Sac_fn_meta, expop) -> [mkStmt (Return(expop, loc))]
        end
  in
  
  (* Elaborate a basic block *)
  let elab_basic_block 
      (bb_stmt_map:stmt Sac_bb_map.t) 
      (bb:sac_basic_block) 
      : stmt 
      =
    begin
      let stmt = Sac_bb_map.find bb bb_stmt_map in
      let instrs = elab_instrs bb.bb_instrs in
      let instrs_stmt = mkStmt (Instr instrs) in
      let term_stmts = elab_term bb_stmt_map bb.bb_term in
      let block = mkBlock (instrs_stmt::term_stmts) in
      stmt.skind <- Block block ;      
      stmt
    end 
  in
  
  (* Elaborate a CFG. *)
  let elab_cfg (cfg:sac_cfg) : block =

(*
    let bb_stmt_map = Slime_cfg.bb_stmt_map cfg in
    let bb0_stmt = (elab_basic_block bb_stmt_map cfg.cfg_bb0) in
    let bbs_stmts = Sac_bb_set.fold 
      (fun bb stmts -> (elab_basic_block bb_stmt_map bb)::stmts) 
      (Sac_bb_set.remove cfg.cfg_bb0 cfg.cfg_bbs) []
    in
    (mkBlock (bb0_stmt :: bbs_stmts))
*)

    let bb_stmt_map = Slime_cfg.bb_stmt_map cfg in
    
    let bbs_stmts = Sac_bb_set.fold 
      (fun bb stmts -> (elab_basic_block bb_stmt_map bb)::stmts) 
      cfg.cfg_bbs []
    in
    
    let entry_stmt = 
      let bb0_stmt = Sac_bb_map.find cfg.cfg_bb0 bb_stmt_map in
      mkStmt (Goto (ref (bb0_stmt), locUnknown))
    in

    (mkBlock (entry_stmt :: bbs_stmts))


  in  

  begin 
    match (fn.fn_source, fn.fn_kind) with
        
      (* Self-adjusting function. *)
      (* BUG (seperate compilation): This pattern matches functions we
         want to elaborate, even if we have only seen their prototypes in
         the source program -- i.e., even for those for which we have a
         "dummy cfg" only. *)
      | (_, Sac_fn_self ob) -> begin
          (* Update the return-type. *)
          fn.fn_varinfo.vtype <-
            (match fn.fn_varinfo.vtype with
                 TFun(voidType, formal_types, false, attr) ->
                   TFun(TPtr(api_typ "closure_t", []), formal_types, false, attr)
               | _ -> raise Hell);          
          
          (* Update the locals -- use the local set from the normalized function *)
          fn.fn_fundec.slocals <- fn.fn_normal.slocals ;
          
          (* Update the function body *)
          fn.fn_fundec.sbody <- elab_cfg fn.fn_cfg ;
        end
          
      | (Some _, Sac_fn_init als) -> 
          (* Update the function body *)
          fn.fn_fundec.sbody <- elab_cfg fn.fn_cfg ;
          
          (*
          (Sac_al_set.iter 
            (fun al -> al.al_blockinfo.bl_aux_init <- make_block_aux_init al)
            als)
          *)
               
      | (Some _, Sac_fn_meta) -> 
          (* Update the function body *)
          fn.fn_fundec.sbody <- elab_cfg fn.fn_cfg ;

      | (None, _) -> ()
  end

(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
(* Elaborates all the self-adjusting functions currently part of the
   program -- this should only be performed once all the
   self-adjusting functions have been added to the program. *)
let elaborate_program _ = begin
  debug "elaborate_program:" ;    
  
  debug "elaborate_program: (preprocessing ...)" ;
  (Sac_fn_set.iter preprocess_function (!sac_functions)) ;

  debug "elaborate_program: (elaborating traceobjs ...):" ;
  (Sac_fn_set.iter elaborate_traceobjs (!sac_functions)) ;
  
  debug "elaborate_program: (creating callgraph ...)" ;
  let cg = Slime_callgraph.create (!sac_functions) in
  
  debug "elaborate_program: (elaborating functions ...):" ;
  (Sac_fn_set.iter (elaborate_function cg) (!sac_functions)) ;

  debug "elaborate_program: (elaborating verifiers ...):" ;
  (Sac_fn_set.iter (elaborate_verifier) (!sac_functions)) ;
  
  debug "elaborating program: done." ;
end

(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
(* Elaborate Function Prototypes *)

let elaborated_prototypes = ref Sac_fn_set.empty
let elaborated_definitions = ref Sac_fn_set.empty

let rec elaborate_prototype (fn:sac_function) (loc:location) : global list =  
  let fn_name = fn.fn_varinfo.vname in
  if (Sac_fn_set.mem fn !elaborated_prototypes) 
  then [GText (sprintf "/* skipping-elaboration-proto:%s */" fn_name)]
  else 
    let _ = elaborated_prototypes := 
      (Sac_fn_set.add fn !elaborated_prototypes) 
    in
    match fn.fn_kind with
        (* Self-adjusting function. *)
        (* Elaborate the declaration of a self-adjusting function --
           this produces two things for the function: (1) a typedef for
           closures of the function (2) a function prototype for
           creating these closures *)
      | Sac_fn_self ob -> 
          GText (sprintf "/* begin-elaborated-proto:%s (self-adjusting) */" fn_name)
          :: (verifier_prototype fn)
          :: (closure_prototype fn ob)
          @[ GText (sprintf "/* end-elaborated-proto:%s (self-adjusting) */" fn_name) ]
          
      (* Initialization function *)
      | Sac_fn_init als ->
          let als = Sac_al_set.elements als in
          GText (sprintf "/* begin-elaborated-proto:%s (init) */" fn_name)
          :: (verifier_prototype fn)
          :: (List.flatten (List.map block_prototype als))
          @[ GText (sprintf "/* end-elaborated-proto:%s (init) */" fn_name) ]
            
      (* Meta function *)
      | Sac_fn_meta ->
          [ GText (sprintf "/* begin-elaborated-proto:%s (meta) */" fn_name) ;
            (verifier_prototype fn) ;
            GVarDecl (fn.fn_varinfo, loc) ;
            GText (sprintf "/* end-elaborated-proto:%s (meta) */" fn_name) ]
            
(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
(* Elaborate Function Definitions *)

let rec elaborate_definition (fn:sac_function) (loc:location) : global list =
  let fn_name = fn.fn_varinfo.vname in
  let fn_kind = (match fn.fn_kind with 
                   | Sac_fn_self _ -> "self-adjusting"
                   | Sac_fn_meta   -> "meta"
                   | Sac_fn_init _ -> "init") in
  
  let _ = elaborated_definitions := 
    (Sac_fn_set.add fn !elaborated_definitions) in

  let text_begin = 
    GText (sprintf "/* begin-elaborated-def:%s (%s) */" fn_name fn_kind) in    

  let elaborated_definition = 
    match fn.fn_kind with      
        (* Self-adjusting function. *)
        (* Elaborate the definition of a self-adjusting function -- this
           produces the elaborated function body as well as other
           boilerplate code we created for the function *)
      | Sac_fn_self ob -> 
          (closure_definition fn ob)
          
      (* Initialization function *)
      (* Elaborate the definitions required by each associated allocation *)
      | Sac_fn_init als -> 
          let als = Sac_al_set.elements als in
          (fn.fn_globs)
          @ [GFun (fn.fn_fundec, locUnknown)]
          @ (List.flatten (List.map block_definition als))
            
      (* Meta function *)
      | Sac_fn_meta ->
          (fn.fn_globs)
          @ [GFun(fn.fn_fundec, loc)]
  in

  let text_end = 
    GText (sprintf "/* end-elaborated-def:%s (%s) */" fn_name fn_kind) in
  
  [text_begin]
  @ [verifier_definition fn]
  @ (elaborate_prototype fn loc)
  @ (elaborated_definition)
  @ [text_end]

(* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - *)
(* Remaining Definitions --- definitions that are needed but haven't
   been yet elaborated by elaborate_definition --- these all correspond
   to external allocation initialization functions for which we only have
   the prototype (but no original definition) *)    

let rec elaborate_remaining_definitions () : global list =
  let elaborate_remaining_definition (fn : sac_function) : global list =
    match fn.fn_kind with
      | Sac_fn_init als -> 
          let fn_name = fn.fn_varinfo.vname in
          let als = Sac_al_set.elements als in
          List.flatten (List.map block_definition als)

      | _ -> 
          let _ = ignore (no_out "remaining def: %s" (fn.fn_varinfo.vname)) in
          []

  in
  let first = GText ("/* begin-remaining-definitions */") in
  let last =  GText ("/* end-remaining-definitions */") in
  
  let required_definitions = 
    Sac_fn_set.diff 
      !elaborated_prototypes 
      !elaborated_definitions
  in

  first 
  :: (List.flatten 
        (Sac_fn_set.fold 
           (fun fn globs -> (elaborate_remaining_definition fn) :: globs)
           required_definitions [[last]]))
