#!/usr/bin/env perl
use strict;
use File::Basename;

my $SRCDIR=$ENV{SRCDIR} or die;
my $TMPSRCDIR=$ENV{TMPSRCDIR} or die;

my $MAIN_C=code("main.template.c");
    
## Give a stacktrace for dies
# use Carp;
# $SIG{ __DIE__ } = sub { Carp::confess( @_ ) };

sub resolve_file {
    my $file = shift;
    my @paths = split("\n", `find $SRCDIR -name $file`);
    my $paths = @paths;
    if($paths == 0) {
        die "Error: `$file' not found, stopped";
    }
    elsif($paths == 1) {
        my $resolved = @paths[0];
        return $resolved;
    }
    else {
        die ("Error: `$file' is not a unique filename\n",
             "Possible matches: ", (join ", ", @paths), "\n",
             "Stopped");
    }
}

sub code_resolve {
    my $code = shift;
    my @resolved_code = map {resolve_file($_)} @$code;
    return \@resolved_code;
}

sub code {
    return code_union (code_resolve [@_]);
}

sub code_union {
    my @union;
    my %seen;
    while(my $code = shift) {
        foreach my $file (@$code) {
            next if $seen{$file}++;
            push @union, $file;
        }
    }
    return \@union;
}

sub code_filter {
    my $suff = shift;
    my $code = shift;
    my @filtered = ();
    
    foreach my $file (@$code) {
        if($file =~ /\.$suff$/) {
            push @filtered, $file;
        }
    }

    return \@filtered;
}


sub make_main_c {
    my $test_app_syms = shift;
    my $main_c_in = shift;
    my $main_c_out = shift;
    
    die unless @$test_app_syms;
    die unless -r $main_c_in;
    die unless -d dirname($main_c_out);

    my $TEST_APP_EXTERNS = 
        join "", (map {"extern test_app_t $_;\n"} @$test_app_syms);

    my $TEST_APP_LIST =
        join ",\n", (map {"&$_"} @$test_app_syms);

    my $TEST_APP_DEFAULT = "NULL";    
    $TEST_APP_DEFAULT = "&@$test_app_syms[0]" 
        if scalar @$test_app_syms == 1;
       
    my $text = `cat $main_c_in`;
    
    $text =~ s/TEST_APP_EXTERNS/$TEST_APP_EXTERNS/g;
    $text =~ s/TEST_APP_LIST/$TEST_APP_LIST/g;
    $text =~ s/TEST_APP_DEFAULT/$TEST_APP_DEFAULT/g;

    open OUT, ">$main_c_out";
    
    print OUT "/* DO NOT EDIT THIS FILE.\n";
    print OUT " * It was autogenerated on ",`date`;
    print OUT " * from $main_c_in.\n";
    print OUT " */\n\n";
    print OUT $text;
    close OUT;    
}

sub apptest {
    my $name = shift;
    my $test_app_sym = shift;
    my $app_code = shift;
    my $test_code = shift;
    
    my $test_app = { 
        name => $name,
        test_app_sym => $test_app_sym,
        app_code => $app_code,
        test_code => $test_code,
        bin => ("\$(BINDIR)/"."test-".$name),
    };

    return $test_app;
}

## Pretty-printing:
sub apptest_pp {
    my $apptest = shift;
    
    print "{\n";
    print "\tname => ", $apptest->{name}, "\n";
    print "\tapp_code => [", (join ", ", @{$apptest->{app_code}}), "]\n";
    print "\ttest_code => [", (join ", ", @{$apptest->{test_code}}), "]\n";
    print "\tbin => ", $apptest->{bin}, "\n";
    print "}\n";
}

## Figure out which directories contain header files that are apart of
## the apptest code-- we use -I arguments (to cilly and gcc) for each
## such directory so that the programmer can omit absolute/relative
## paths when they #include these headers in their C code.
sub apptest_include_dirs {
    my $apptest = shift;    
    my @dirs = ();
    my %seen = ();

    foreach my $code (@{$apptest->{app_code}},
                      @{$apptest->{test_code}}) 
    {
        if($code =~ /\.h$/) {
            my $dir = dirname($code);
            next if $seen{$dir}++;
            push @dirs, $dir;
        }
    }
    return \@dirs;
}

sub apptest_make_rule {
    my $apptest = shift;    
    
    my $test_app_sym = $apptest->{test_app_sym};

    my $main_c_in = @$MAIN_C[0];
    my $main_c_out = "$TMPSRCDIR/$test_app_sym.main.c";
    
    make_main_c([$test_app_sym], $main_c_in, $main_c_out);

    my $all_code = code_union($apptest->{app_code}, 
                              $apptest->{test_code}, 
                              code(basename($main_c_out)));

    my $all_c_code = code_filter("c", $all_code);
    
    my $app_c_code = code_filter("c", $apptest->{app_code});
    
    my $include_dirs = apptest_include_dirs($apptest);

    ## The makefile rule format: "TARGET: DEPENDENCIES"
    print ($apptest->{bin}, 
           ": ", 
           "\$(CEALC) \$(LIBCEAL) ",
           (join " ", @$all_code), 
           "\n");

    if(0) {
     ## Time the compilation (of the application code only):
    print ("\t","gcc ",
           "\$(CFLAGS) ",
           (map {"-I$_ "} @$include_dirs),            
           (map {"$_ "} @$all_c_code),
           "-o ", "$apptest->{bin}.tmp ",
           "\$(LIBCEAL) ", ## Has to be last?
           "\n");
}

    ## Compile application and testing code with CEALC:
    print ("\t", "\$(CEALC) ",
           "\$(CFLAGS) \$(CILLY_FLAGS) \$(CEALC_FLAGS) ",
           (map {"-I$_ "} @$include_dirs),            
           (map {"$_ "} @$all_c_code),
           "-o ", $apptest->{bin},
           "\n");

    print "\n";
}

sub apptest_makefile {
    my $apptests = shift;
    
    print "## DO NOT EDIT.\n";
    print "## Automatically generated on ",`date | tr -d '\n\t'`,"\n";
    print "## Generated from: $0\n";
    print "\n";
    
    print "export CILLY\n";
    print "export CILLY_FLAGS\n";
    print "export LIBCEAL\n";
    print "export CEALC_FLAGS\n";
    print "\n";
    
    ## Create the 'all' rule:
    my @all_bins = map {($_->{bin})} @$apptests;
    print "all: ", (join " ", @all_bins), "\n\n";
    
    ## Create the other rules:
    map {apptest_make_rule($_)} @$apptests;
}

die ("$0: Error: Can\'t find directory: $SRCDIR\n",
     "You are running this in the wrong directory.")
    unless -d $SRCDIR;

1;
